==> klayout-0.9.tar.gz: klayout-0.9/src/layDitherPattern.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "layDitherPattern.h"

#include <ctype.h>

namespace lay
{

// ---------------------------------------------------------------------
//  The standard dither pattern

static const char *dither_strings [] = {

  // 0: solid
  "solid",
  "*",

  // 1: hollow
  "hollow",
  ".", 

  // 2: dotted
  "dotted",
  "*.\n"
  ".*",

  // 3: coarsly dotted
  "coarsly dotted",
  "*...\n"
  "....\n"
  "..*.\n"
  "....",

  // 4: left-hatched
  "left-hatched",
  "*...\n"
  ".*..\n"
  "..*.\n"
  "...*",

  // 5: lightly left-hatched
  "lightly left-hatched",
  "*.......\n"
  ".*......\n"
  "..*.....\n"
  "...*....\n"
  "....*...\n"
  ".....*..\n"
  "......*.\n"
  ".......*",

  // 6: strongly left-hatched dense
  "strongly left-hatched dense",
  "**..\n"
  ".**.\n"
  "..**\n"
  "*..*",

  // 7: strongly left-hatched sparse
  "strongly left-hatched sparse",
  "**......\n"
  ".**.....\n"
  "..**....\n"
  "...**...\n"
  "....**..\n"
  ".....**.\n"
  "......**\n"
  "*......*",

  // 8: right-hatched
  "right-hatched",
  "*...\n"
  "...*\n"
  "..*.\n"
  ".*..",

  // 9: lightly right-hatched
  "lightly right-hatched",
  "*.......\n"
  ".......*\n"
  "......*.\n"
  ".....*..\n"
  "....*...\n"
  "...*....\n"
  "..*.....\n"
  ".*......",

  // 10: strongly right-hatched dense
  "strongly right-hatched dense",
  "**..\n"
  "*..*\n"
  "..**\n"
  ".**.",

  // 11: strongly right-hatched sparse
  "strongly right-hatched sparse",
  "**......\n"
  "*......*\n"
  "......**\n"
  ".....**.\n"
  "....**..\n"
  "...**...\n"
  "..**....\n"
  ".**.....",

  // 12: cross-hatched
  "cross-hatched",
  "*...\n"
  ".*.*\n"
  "..*.\n"
  ".*.*",

  // 13: lightly cross-hatched
  "lightly cross-hatched",
  "*.......\n"
  ".*.....*\n"
  "..*...*.\n"
  "...*.*..\n"
  "....*...\n"
  "...*.*..\n"
  "..*...*.\n"
  ".*.....*",

  // 14: checkerboard 2px
  "checkerboard 2px",
  "**..\n"
  "**..\n"
  "..**\n"
  "..**",

  // 15: strongly cross-hatched sparse
  "strongly cross-hatched sparse",
  "**......\n"
  "***....*\n"
  "..**..**\n"
  "...****.\n"
  "....**..\n"
  "...****.\n"
  "..**..**\n"
  "***....*"

};


// ---------------------------------------------------------------------
//  DitherPatternInfo implementation

DitherPatternInfo::DitherPatternInfo ()
  : m_order_index (0)
{
  memset (m_pattern, 0, sizeof (m_pattern));
}
  
DitherPatternInfo::DitherPatternInfo (const DitherPatternInfo &d)
  : m_order_index (d.m_order_index), m_name (d.m_name)
{
  memcpy (m_pattern, d.m_pattern, sizeof (m_pattern));
}
  
DitherPatternInfo &
DitherPatternInfo::operator= (const DitherPatternInfo &d)
{
  if (&d != this) {
    m_order_index = d.m_order_index;
    m_name = d.m_name;
    memcpy (m_pattern, d.m_pattern, sizeof (m_pattern));
  }
  return *this;
}

bool 
DitherPatternInfo::operator== (const DitherPatternInfo &d) const
{
  for (unsigned int i = 0; i < sizeof (m_pattern) / sizeof (m_pattern [0]); ++i) {
    if (m_pattern [i] != d.m_pattern [i]) {
      return false;
    }
  }
  return m_name == d.m_name && m_order_index == d.m_order_index;
}

void 
DitherPatternInfo::set_pattern (const uint32_t *pt, unsigned int w, unsigned int h) 
{
  if (w == 0) {
    w = 1;
  } else if (w >= 32) {
    w = 32;
  } 

  if (h == 0) {
    h = 1;
  } else if (h >= 32) {
    h = 32;
  } 

  for (unsigned int n = 0; n < 64; ++n) {

    uint32_t wd;
    if (w < 32) {
      wd = pt[n % h] & ((1 << w) - 1);
    } else {
      wd = pt[n % h];
    }

    unsigned int wl = w;
    while (wl < 32) {
      wd = wd | (wd << wl);
      wl *= 2;
    }

    if (m_pattern [n] != wd) {
      m_pattern [n] = wd;
    }

  }

}

std::string
DitherPatternInfo::to_string () const
{
  std::string res;
  
  for (unsigned int i = 0; i < 32; ++i) {
    res += "          ";
    for (unsigned int j = 0; j < 32; ++j) {
      if ((m_pattern [32 - i - 1] & (1 << j)) != 0) {
        res += "*";
      } else {
        res += ".";
      }
    }
    res += "\n";
  }

  return res;
}

std::vector <std::string>
DitherPatternInfo::to_strings () const
{
  std::vector <std::string> res;
  
  for (unsigned int i = 0; i < 32; ++i) {
    std::string r;
    for (unsigned int j = 0; j < 32; ++j) {
      if ((m_pattern [32 - i - 1] & (1 << j)) != 0) {
        r += "*";
      } else {
        r += ".";
      }
    }
    res.push_back (r);
  }

  return res;
}

static const char *uint_from_string (const char *s, uint32_t &w)
{
  while (*s && isspace (*s)) {
    ++s;
  }

  w = 0;

  const char *s0 = s;
  const char *s1 = s0;
  while (*s1 >= ' ') {
    ++s1;
  }

  unsigned int b = 1;
  while (b > 0) {
    if (*s < ' ') {
      s = s0;
    } 
    if (*s++ == '*') {
      w |= b;
    }
    b <<= 1;
  }

  return s1;
}

void
DitherPatternInfo::from_strings (const std::vector<std::string> &strv) 
{
  for (unsigned int l = 0; l < 64; ++l) {
    m_pattern[l] = 0;
    if (strv.size () > 0) {
      const char *s = strv [l % strv.size ()].c_str ();
      s = uint_from_string (s, m_pattern [64 - l - 1]);
    }
  }
}

void
DitherPatternInfo::from_string (const std::string &cstr) 
{
  const char *ds = cstr.c_str ();
  const char *s = ds;

  for (unsigned int l = 0; l < 64; ++l) {

    m_pattern[l] = 0;

    while (*s && isspace (*s)) {
      ++s;
    }
    if (! *s) {
      s = ds;
    }
    
    s = uint_from_string (s, m_pattern [64 - l - 1]);

  }
}

// ---------------------------------------------------------------------
//  DitherPattern implementation

struct ReplaceDitherPatternOp
  : public db::Op
{
  ReplaceDitherPatternOp (unsigned int i, const DitherPatternInfo &o, const DitherPatternInfo &n) 
    : db::Op (), index (i), m_old (o), m_new (n) 
  { }

  unsigned int index;
  DitherPatternInfo m_old, m_new;
};

DitherPattern::DitherPattern ()
  : QObject (), db::Object (0)
{
  for (unsigned int d = 0; d < sizeof (dither_strings) / sizeof (dither_strings [0]); d += 2) {
    m_pattern.push_back (DitherPatternInfo ());
    m_pattern.back ().set_name (dither_strings [d]);
    m_pattern.back ().from_string (dither_strings [d + 1]);
  }
}

DitherPattern::~DitherPattern ()
{
  //  .. nothing yet ..
}

DitherPattern::DitherPattern (const DitherPattern &p)
{
  m_pattern = p.m_pattern;
}

DitherPattern & 
DitherPattern::operator= (const DitherPattern &p)
{
  if (this != &p) {
    unsigned int i;
    for (i = 0; i < p.count (); ++i) {
      replace_pattern (i, p.begin () [i]);
    }
    for ( ; i < count (); ++i) {
      replace_pattern (i, DitherPatternInfo ());
    }
  }
  return *this;
}

const uint32_t *
DitherPattern::pattern (unsigned int i) const
{
  if (i < count ()) {
    return m_pattern [i].pattern ();
  } else {
    return m_pattern [0].pattern (); //  empty
  }
}

void 
DitherPattern::replace_pattern (unsigned int i, const DitherPatternInfo &p)
{
  bool chg = false;

  while (i >= count ()) {
    m_pattern.push_back (DitherPatternInfo ());
    chg = true;
  }

  if (m_pattern [i] != p) {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new ReplaceDitherPatternOp (i, m_pattern [i], p));
    }
    m_pattern [i] = p;
    chg = true;
  }

  //  if something has changed emit the signal
  if (chg) {
    emit changed ();
  }
}

DitherPattern::iterator 
DitherPattern::begin_custom () const 
{
  return m_pattern.begin () + sizeof (dither_strings) / sizeof (dither_strings [0]) / 2;
}

void 
DitherPattern::undo (db::Op *op) throw ()
{
  const ReplaceDitherPatternOp *rop = dynamic_cast <const ReplaceDitherPatternOp *> (op);
  if (rop) {
    replace_pattern (rop->index, rop->m_old);
  }
}

void 
DitherPattern::redo (db::Op *op) throw ()
{
  const ReplaceDitherPatternOp *rop = dynamic_cast <const ReplaceDitherPatternOp *> (op);
  if (rop) {
    replace_pattern (rop->index, rop->m_new);
  }
}

}

==> klayout-0.10.tar.gz: klayout-0.10/src/layDitherPattern.cc <==

/*

  KLayout Layout Viewer
  Copyright (C) 2006 Matthias Koefferlein

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

*/



#include "layDitherPattern.h"

#include <ctype.h>

namespace lay
{

// ---------------------------------------------------------------------
//  The standard dither pattern

static const char *dither_strings [] = {

  // 0: solid
  "solid",
  "*",

  // 1: hollow
  "hollow",
  ".", 

  // 2: dotted
  "dotted",
  "*.\n"
  ".*",

  // 3: coarsly dotted
  "coarsly dotted",
  "*...\n"
  "....\n"
  "..*.\n"
  "....",

  // 4: left-hatched
  "left-hatched",
  "*...\n"
  ".*..\n"
  "..*.\n"
  "...*",

  // 5: lightly left-hatched
  "lightly left-hatched",
  "*.......\n"
  ".*......\n"
  "..*.....\n"
  "...*....\n"
  "....*...\n"
  ".....*..\n"
  "......*.\n"
  ".......*",

  // 6: strongly left-hatched dense
  "strongly left-hatched dense",
  "**..\n"
  ".**.\n"
  "..**\n"
  "*..*",

  // 7: strongly left-hatched sparse
  "strongly left-hatched sparse",
  "**......\n"
  ".**.....\n"
  "..**....\n"
  "...**...\n"
  "....**..\n"
  ".....**.\n"
  "......**\n"
  "*......*",

  // 8: right-hatched
  "right-hatched",
  "*...\n"
  "...*\n"
  "..*.\n"
  ".*..",

  // 9: lightly right-hatched
  "lightly right-hatched",
  "*.......\n"
  ".......*\n"
  "......*.\n"
  ".....*..\n"
  "....*...\n"
  "...*....\n"
  "..*.....\n"
  ".*......",

  // 10: strongly right-hatched dense
  "strongly right-hatched dense",
  "**..\n"
  "*..*\n"
  "..**\n"
  ".**.",

  // 11: strongly right-hatched sparse
  "strongly right-hatched sparse",
  "**......\n"
  "*......*\n"
  "......**\n"
  ".....**.\n"
  "....**..\n"
  "...**...\n"
  "..**....\n"
  ".**.....",

  // 12: cross-hatched
  "cross-hatched",
  "*...\n"
  ".*.*\n"
  "..*.\n"
  ".*.*",

  // 13: lightly cross-hatched
  "lightly cross-hatched",
  "*.......\n"
  ".*.....*\n"
  "..*...*.\n"
  "...*.*..\n"
  "....*...\n"
  "...*.*..\n"
  "..*...*.\n"
  ".*.....*",

  // 14: checkerboard 2px
  "checkerboard 2px",
  "**..\n"
  "**..\n"
  "..**\n"
  "..**",

  // 15: strongly cross-hatched sparse
  "strongly cross-hatched sparse",
  "**......\n"
  "***....*\n"
  "..**..**\n"
  "...****.\n"
  "....**..\n"
  "...****.\n"
  "..**..**\n"
  "***....*"

};


// ---------------------------------------------------------------------
//  DitherPatternInfo implementation

DitherPatternInfo::DitherPatternInfo ()
  : m_order_index (0)
{
  memset (m_pattern, 0, sizeof (m_pattern));
}
  
DitherPatternInfo::DitherPatternInfo (const DitherPatternInfo &d)
  : m_order_index (d.m_order_index), m_name (d.m_name)
{
  memcpy (m_pattern, d.m_pattern, sizeof (m_pattern));
}
  
DitherPatternInfo &
DitherPatternInfo::operator= (const DitherPatternInfo &d)
{
  if (&d != this) {
    m_order_index = d.m_order_index;
    m_name = d.m_name;
    memcpy (m_pattern, d.m_pattern, sizeof (m_pattern));
  }
  return *this;
}

bool 
DitherPatternInfo::operator== (const DitherPatternInfo &d) const
{
  for (unsigned int i = 0; i < sizeof (m_pattern) / sizeof (m_pattern [0]); ++i) {
    if (m_pattern [i] != d.m_pattern [i]) {
      return false;
    }
  }
  return m_name == d.m_name && m_order_index == d.m_order_index;
}

void 
DitherPatternInfo::set_pattern (const uint32_t *pt, unsigned int w, unsigned int h) 
{
  if (w == 0) {
    w = 1;
  } else if (w >= 32) {
    w = 32;
  } 

  if (h == 0) {
    h = 1;
  } else if (h >= 32) {
    h = 32;
  } 

  for (unsigned int n = 0; n < 64; ++n) {

    uint32_t wd;
    if (w < 32) {
      wd = pt[n % h] & ((1 << w) - 1);
    } else {
      wd = pt[n % h];
    }

    unsigned int wl = w;
    while (wl < 32) {
      wd = wd | (wd << wl);
      wl *= 2;
    }

    m_pattern [n] = wd;

  }

}

std::string
DitherPatternInfo::to_string () const
{
  std::string res;
  
  for (unsigned int i = 0; i < 32; ++i) {
    res += "          ";
    for (unsigned int j = 0; j < 32; ++j) {
      if ((m_pattern [32 - i - 1] & (1 << j)) != 0) {
        res += "*";
      } else {
        res += ".";
      }
    }
    res += "\n";
  }

  return res;
}

std::vector <std::string>
DitherPatternInfo::to_strings () const
{
  std::vector <std::string> res;
  
  for (unsigned int i = 0; i < 32; ++i) {
    std::string r;
    for (unsigned int j = 0; j < 32; ++j) {
      if ((m_pattern [32 - i - 1] & (1 << j)) != 0) {
        r += "*";
      } else {
        r += ".";
      }
    }
    res.push_back (r);
  }

  return res;
}

static const char *uint_from_string (const char *s, uint32_t &w)
{
  while (*s && isspace (*s)) {
    ++s;
  }

  w = 0;

  const char *s0 = s;
  const char *s1 = s0;
  while (*s1 >= ' ') {
    ++s1;
  }

  unsigned int b = 1;
  while (b > 0) {
    if (*s < ' ') {
      s = s0;
    } 
    if (*s++ == '*') {
      w |= b;
    }
    b <<= 1;
  }

  return s1;
}

void
DitherPatternInfo::from_strings (const std::vector<std::string> &strv) 
{
  for (unsigned int l = 0; l < 64; ++l) {
    m_pattern[64 - l - 1] = 0;
    if (strv.size () > 0) {
      const char *s = strv [l % strv.size ()].c_str ();
      s = uint_from_string (s, m_pattern [64 - l - 1]);
    }
  }
}

void
DitherPatternInfo::from_string (const std::string &cstr) 
{
  const char *ds = cstr.c_str ();
  const char *s = ds;

  for (unsigned int l = 0; l < 64; ++l) {

    m_pattern[64 - l - 1] = 0;

    while (*s && isspace (*s)) {
      ++s;
    }
    if (! *s) {
      s = ds;
    }
    
    s = uint_from_string (s, m_pattern [64 - l - 1]);

  }
}

// ---------------------------------------------------------------------
//  DitherPattern implementation

struct ReplaceDitherPatternOp
  : public db::Op
{
  ReplaceDitherPatternOp (unsigned int i, const DitherPatternInfo &o, const DitherPatternInfo &n) 
    : db::Op (), index (i), m_old (o), m_new (n) 
  { }

  unsigned int index;
  DitherPatternInfo m_old, m_new;
};

DitherPattern::DitherPattern ()
  : QObject (), db::Object (0)
{
  for (unsigned int d = 0; d < sizeof (dither_strings) / sizeof (dither_strings [0]); d += 2) {
    m_pattern.push_back (DitherPatternInfo ());
    m_pattern.back ().set_name (dither_strings [d]);
    m_pattern.back ().from_string (dither_strings [d + 1]);
  }
}

DitherPattern::~DitherPattern ()
{
  //  .. nothing yet ..
}

DitherPattern::DitherPattern (const DitherPattern &p)
{
  m_pattern = p.m_pattern;
}

DitherPattern & 
DitherPattern::operator= (const DitherPattern &p)
{
  if (this != &p) {
    unsigned int i;
    for (i = 0; i < p.count (); ++i) {
      replace_pattern (i, p.begin () [i]);
    }
    for ( ; i < count (); ++i) {
      replace_pattern (i, DitherPatternInfo ());
    }
  }
  return *this;
}

const uint32_t *
DitherPattern::pattern (unsigned int i) const
{
  unsigned int off = (i >> 16) % 32;
  i &= 0xffff;
  if (i < count ()) {
    return m_pattern [i].pattern () + off;
  } else {
    return m_pattern [1].pattern (); //  empty
  }
}

void 
DitherPattern::replace_pattern (unsigned int i, const DitherPatternInfo &p)
{
  bool chg = false;

  while (i >= count ()) {
    m_pattern.push_back (DitherPatternInfo ());
    chg = true;
  }

  if (m_pattern [i] != p) {
    if (manager () && manager ()->transacting ()) {
      manager ()->queue (this, new ReplaceDitherPatternOp (i, m_pattern [i], p));
    }
    m_pattern [i] = p;
    chg = true;
  }

  //  if something has changed emit the signal
  if (chg) {
    emit changed ();
  }
}

DitherPattern::iterator 
DitherPattern::begin_custom () const 
{
  return m_pattern.begin () + sizeof (dither_strings) / sizeof (dither_strings [0]) / 2;
}

void 
DitherPattern::undo (db::Op *op) throw ()
{
  const ReplaceDitherPatternOp *rop = dynamic_cast <const ReplaceDitherPatternOp *> (op);
  if (rop) {
    replace_pattern (rop->index, rop->m_old);
  }
}

void 
DitherPattern::redo (db::Op *op) throw ()
{
  const ReplaceDitherPatternOp *rop = dynamic_cast <const ReplaceDitherPatternOp *> (op);
  if (rop) {
    replace_pattern (rop->index, rop->m_new);
  }
}

}
